#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "gmp.h"
#define showz(x,a,b) {printf(#x"(%d,%d)=", a, b); mpz_out_str(stdout, 10, x); printf("\n"); }
#define A   13591409
#define B   545140134
#define C   640320
#define D   12
#define BITS_PER_DIGIT   3.32192809488736234787
#define DIGITS_PER_ITER   14.181
#define DOUBLE_PREC         53
mpf_t t0, t1, t2;

void
my_sqrt_ui (mpf_t r, unsigned x)
{
  unsigned prec, bits, prec0;

  prec0 = mpf_get_prec (r);

  if (prec0 <= DOUBLE_PREC)
    {
      mpf_set_d (r, sqrt (x));
      return;
    }

  bits = 0;
  for (prec = prec0; prec > DOUBLE_PREC;)
    {
      int bit = prec & 1;
      prec = (prec + bit) / 2;
      bits = bits * 2 + bit;
    }

  mpf_set_prec_raw (t1, DOUBLE_PREC);
  mpf_set_d (t1, 1 / sqrt (x));

  while (prec < prec0)
    {
      prec *= 2;
      if (prec < prec0)
	{
	  /* t1 = t1+t1*(1-x*t1*t1)/2; */
	  mpf_set_prec_raw (t2, prec);
	  mpf_mul (t2, t1, t1);
	  mpf_mul_ui (t2, t2, x);
	  mpf_ui_sub (t2, 1, t2);
	  mpf_set_prec_raw (t2, prec / 2);
	  mpf_div_2exp (t2, t2, 1);
	  mpf_mul (t2, t2, t1);
	  mpf_set_prec_raw (t1, prec);
	  mpf_add (t1, t1, t2);
	}

      else
	{
	  prec = prec0;
	  /* t2=x*t1, t1 = t2+t1*(x-t2*t2)/2; */
	  mpf_set_prec_raw (t2, prec / 2);
	  mpf_mul_ui (t2, t1, x);
	  mpf_mul (t0, t2, t2);
	  mpf_ui_sub (t0, x, t0);
	  mpf_mul (t1, t1, t0);
	  mpf_div_2exp (t1, t1, 1);
	  mpf_add (r, t1, t2);
	  break;
	}

      prec -= (bits & 1);
      bits /= 2;
    }
}

void
my_div (mpf_t r, mpf_t y, mpf_t x)
{
  unsigned prec, bits, prec0;

  prec0 = mpf_get_prec (r);

  if (prec0 <= DOUBLE_PREC)
    {
      mpf_set_d (r, mpf_get_d (y) / mpf_get_d (x));
      return;
    }

  bits = 0;
  for (prec = prec0; prec > DOUBLE_PREC;)
    {
      int bit = prec & 1;
      prec = (prec + bit) / 2;
      bits = bits * 2 + bit;
    }

  mpf_set_prec_raw (t1, DOUBLE_PREC);
  mpf_set_d (t1, 1 / mpf_get_d (x));

  while (prec < prec0)
    {
      prec *= 2;
      if (prec < prec0)
	{
	  /* t1 = t1+t1*(1-x*t1); */
	  mpf_set_prec_raw (t2, prec);
	  mpf_mul (t2, t1, x);
	  mpf_ui_sub (t2, 1, t2);
	  mpf_set_prec_raw (t2, prec);
	  mpf_mul (t2, t2, t1);
	  mpf_set_prec_raw (t1, prec);
	  mpf_add (t1, t1, t2);
	}
      else
	{
	  prec = prec0;
	  /* t2=y*t1, t1 = t2+t1*(y-x*t2); */
	  mpf_set_prec_raw (t2, prec);
	  mpf_mul (t2, t1, y);
	  mpf_mul (t0, x, t2);
	  mpf_sub (t0, y, t0);
	  mpf_mul (t1, t1, t0);
	  mpf_add (r, t1, t2);
	  break;
	}
      prec -= (bits & 1);
      bits /= 2;
    }
}

int out = 0;

bs (mpz_t p, mpz_t q, mpz_t g, unsigned a, unsigned b, unsigned gflag)
{
  mpz_t p2, q2, g2;
  unsigned i;

  switch (b - a)
    {
    case 1:
      /*
         g(b-1,b) = (6b-5)(6b-4)(6b-3)(6b-2)(6b-1)(6b)
         p(b-1,b) = b^3 * C^3 * (3b-2)(3b-1)(3b)
         q(b-1,b) = (-1)^b*g(b-1,b)*(A+Bb).
         Reduced to:
         g(b-1,b) = (6b-5)(2b-1)(6b-1)
         p(b-1,b) = b^3 * C^3 / 24
         q(b-1,b) = (-1)^b*g(b-1,b)*(A+Bb).
       */

      mpz_set_ui (g, 2 * b - 1);
      mpz_mul_ui (g, g, 6 * b - 1);
      mpz_mul_ui (g, g, 6 * b - 5);

      mpz_set_ui (p, b);
      mpz_mul_ui (p, p, b);
      mpz_mul_ui (p, p, b);
      mpz_mul_ui (p, p, (C / 24) * (C / 24));
      mpz_mul_ui (p, p, C * 24);

      mpz_set_ui (q, b);
      mpz_mul_ui (q, q, B);
      mpz_add_ui (q, q, A);
      mpz_mul (q, q, g);
      if (b % 2)
	mpz_neg (q, q);
      break;

    default:
      /*
         p(a,b) = p(a,m) * p(m,b)
         q(a,b) = q(a,m) * p(m,b) + q(m,b) * g(a,m)
         g(a,b) = g(a,m) * g(m,b)
       */

      mpz_init (p2);
      mpz_init (q2);
      mpz_init (g2);

      bs (p, q, g, a, (b + a) / 2, 1);
      bs (p2, q2, g2, (b + a) / 2, b, gflag);

      mpz_mul (p, p, p2);
      mpz_mul (q, q, p2);
      mpz_mul (q2, q2, g);
      mpz_add (q, q, q2);
      if (gflag)
	mpz_mul (g, g, g2);

      mpz_clear (p2);
      mpz_clear (q2);
      mpz_clear (g2);
      break;
    }
  if (out & 2)
    {
      showz (p, a, b);
      showz (q, a, b);
      if (gflag)
	showz (g, a, b);
    }
}

main (int argc, char *argv[])
{
  mpz_t p, q, g;
  mpf_t pi, qi;
  int d, iter, i;
  clock_t begin, mid1, mid2, mid3, mid4, end;

  if (argc < 2)
    {
      printf ("%s  #digits [output]\n"
	      "    output=0, no output\n"
	      "           1, output to stdout\n", argv[0]);
      exit (0);
    }
  d = atoi (argv[1]);
  if (argc > 2)
    out = atoi (argv[2]);

  iter = d / DIGITS_PER_ITER;

  begin = clock ();

  mpz_init (p);
  mpz_init (q);
  mpz_init (g);

  if (iter <= 0)
    {
      mpz_set_ui (p, 1);
      mpz_set_ui (q, 0);
      mpz_set_ui (g, 1);
    }
  else
    {
      bs (p, q, g, 0, iter, 0);
    }

  mid1 = clock ();

  mpf_set_default_prec (d * BITS_PER_DIGIT + 16);

  mpf_init (pi);
  mpf_init (qi);
  mpf_init (t0);
  mpf_init (t1);
  mpf_init (t2);

  /*
     p*(C/D)*sqrt(C)
     pi = -----------------
     (q+A*p)
   */
  mpz_addmul_ui (q, p, A);
  mpz_mul_ui (p, p, C / D);

  mpf_set_z (pi, p);
  mpf_set_z (qi, q);
  mpf_div_2exp (pi, pi, mpz_sizeinbase (q, 2));
  mpf_div_2exp (qi, qi, mpz_sizeinbase (q, 2));

  mid2 = clock ();

  my_sqrt_ui (t0, C);

  mid3 = clock ();

  mpf_mul (pi, pi, t0);

  mid4 = clock ();

  my_div (pi, pi, qi);

  end = clock ();

  if (out & 1)
    {
      printf ("pi(%d,%d)=\n", 0, iter);
      mpf_out_str (stdout, 10, d + 2, pi);
      printf ("\n");
    }
  printf ("Binary splitting     time=%f\n",
	  (double) (mid1 - begin) / CLOCKS_PER_SEC);
  printf ("Integer to float     time=%f\n",
	  (double) (mid2 - mid1) / CLOCKS_PER_SEC);
  printf ("Final square root    time=%f\n",
	  (double) (mid3 - mid2) / CLOCKS_PER_SEC);
  printf ("Final multiplication time=%f\n",
	  (double) (mid4 - mid3) / CLOCKS_PER_SEC);
  printf ("Final division       time=%f\n",
	  (double) (end - mid4) / CLOCKS_PER_SEC);
  printf ("Total                time=%f\n",
	  (double) (end - begin) / CLOCKS_PER_SEC);

  mpz_clear (p);
  mpz_clear (q);
  mpz_clear (g);

  mpf_clear (pi);
  mpf_clear (qi);
}
